From 89ce4ea1cb7bbb809c494cdf49f7a01d780cbb35 Mon Sep 17 00:00:00 2001
From: Alyssa Rosenzweig <alyssa@collabora.com>
Date: Tue, 25 Oct 2022 22:29:31 -0400
Subject: [PATCH] nir/opt_algebraic: Fuse c - a * b to FMA

Algebraically it is clear that

   -(a * b) + c = (-a) * b + c = fma(-a, b, c)

But this is not clear from the NIR

   ('fadd', ('fneg', ('fmul', a, b)), c)

Add rules to handle this case specially. Note we don't necessarily want
to  solve this by pushing fneg into fmul, because the rule opt_algebraic
(not the late part where FMA fusing happens) specifically pulls fneg out
of fmul to push fneg up multiplication chains.

Noticed in the big glmark2 "terrain" shader, which has a cycle count
reduced by 22% on Mali-G57 thanks to having this pattern a ton and being
FMA bound.

BEFORE: 1249 inst, 16.015625 cycles, 16.015625 fma, ... 632 quadwords
AFTER: 997 inst, 12.437500 cycles, .... 504 quadwords

Results on the same shader on AGX are also quite dramatic:

BEFORE: 1294 inst, 8600 bytes, 50 halfregs, ...
AFTER: 1154 inst, 8040 bytes, 50 halfregs, ...

TODO: Do the same thing with fabs and fneg(fabs)?

Signed-off-by: Alyssa Rosenzweig <alyssa@collabora.com>
---
 src/compiler/nir/nir_opt_algebraic.py | 5 +++++
 1 file changed, 5 insertions(+)

diff --git a/src/compiler/nir/nir_opt_algebraic.py b/src/compiler/nir/nir_opt_algebraic.py
index cc55b866657..37e7be4441d 100644
--- a/src/compiler/nir/nir_opt_algebraic.py
+++ b/src/compiler/nir/nir_opt_algebraic.py
@@ -2631,6 +2631,11 @@ late_optimizations = [
    (('~fadd@64', ('fmul(is_only_used_by_fadd)', a, b), c), ('ffma', a, b, c), 'options->fuse_ffma64'),
    (('~fadd@32', ('fmulz(is_only_used_by_fadd)', a, b), c), ('ffmaz', a, b, c), 'options->fuse_ffma32'),
 
+   (('~fadd@16', ('fneg(is_only_used_by_fadd)', ('fmul', a, b)), c), ('ffma', ('fneg', a), b, c), 'options->fuse_ffma16'),
+   (('~fadd@32', ('fneg(is_only_used_by_fadd)', ('fmul', a, b)), c), ('ffma', ('fneg', a), b, c), 'options->fuse_ffma32'),
+   (('~fadd@64', ('fneg(is_only_used_by_fadd)', ('fmul', a, b)), c), ('ffma', ('fneg', a), b, c), 'options->fuse_ffma64'),
+   (('~fadd@32', ('fneg(is_only_used_by_fadd)', ('fmulz', a, b)), c), ('ffmaz', ('fneg', a), b, c), 'options->fuse_ffma32'),
+
    # Subtractions get lowered during optimization, so we need to recombine them
    (('fadd@8', a, ('fneg', 'b')), ('fsub', 'a', 'b'), 'options->has_fsub'),
    (('fadd@16', a, ('fneg', 'b')), ('fsub', 'a', 'b'), 'options->has_fsub'),
-- 
2.35.1

